/*
   +----------------------------------------------------------------------+
   | PHP bcompiler Based on APC Cache                                                 |
   +----------------------------------------------------------------------+
   | Copyright (c) 2000-2001 Community Connect, Inc. (apc_* functions)    |
   | Copyright (c) 1997-2002 The PHP Group (remaining code)               |
   +----------------------------------------------------------------------+
   | This source file is subject to version 2.02 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available at through the world-wide-web at                           |
   | http://www.php.net/license/2_02.txt.                                 |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Authors: Alan Knowles <alan@akbkhome.com> (php module)		  |
   | Daniel Cowgill <dan@mail.communityconnect.com>			  |
   | George Schlossnagle <george@lethargy.org>                            |
   | Mike Bretz <mike@metropolis-ag.de>                                   |
   +----------------------------------------------------------------------+
 */
/* $Id: bcompiler.c,v 1.10 2002/09/07 02:34:30 alan_k Exp $ */
 
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "ext/standard/file.h"
#include "php_bcompiler.h"
#include "zend.h"
 

#include <stdlib.h>
#include <assert.h>

/* Define the header & version */

#define BCOMPILER_VERSION "0.3"
#ifdef PHP_HAVE_STREAMS
#define BCOMPILER_STREAMS "s"
#else
#define BCOMPILER_STREAMS ""
#endif
#define BCOMPILER_MAGIC_HEADER "bcompiler v"  BCOMPILER_VERSION  BCOMPILER_STREAMS

/* Debugging */

#define DEBUGLEVEL 0 
#if DEBUGLEVEL
#define BCOMPILER_DEBUG(args)  (printf("DEBUG: "), printf args)
#else
#define BCOMPILER_DEBUG(args)  
#endif

/* we use globals */
ZEND_DECLARE_MODULE_GLOBALS(bcompiler)



/* {{{ bcompiler_functions[]
 *
 * All the functions in bcompiler
 */
function_entry bcompiler_functions[] = {
        PHP_FE(bcompiler_write_header,           NULL)         
        PHP_FE(bcompiler_write_class,            NULL)         
        PHP_FE(bcompiler_write_footer,           NULL)         
        PHP_FE(bcompiler_write_exe_footer,       NULL)         
        PHP_FE(bcompiler_read,                   NULL)         
        PHP_FE(bcompiler_load,                   NULL)         
        PHP_FE(bcompiler_load_exe,               NULL)         
        PHP_FE(bcompiler_parse_class,            NULL)         
        {NULL, NULL, NULL}	 
};
/* }}} */

/* {{{ bcompiler_module_entry
 */
zend_module_entry bcompiler_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"bcompiler",
	bcompiler_functions,
	PHP_MINIT(bcompiler),
	PHP_MSHUTDOWN(bcompiler),
	PHP_RINIT(bcompiler),		 
	PHP_RSHUTDOWN(bcompiler),	 
	PHP_MINFO(bcompiler),
#if ZEND_MODULE_API_NO >= 20010901
	BCOMPILER_VERSION  BCOMPILER_STREAMS, /* Replace with version number for your extension */
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_BCOMPILER
ZEND_GET_MODULE(bcompiler)
#endif


 
/* {{{ php_bcompiler_init_globals
 */
static void php_bcompiler_init_globals(zend_bcompiler_globals *bcompiler_globals)
{
	
       
        bcompiler_globals->buffer = (char*) emalloc(1);
        bcompiler_globals->stream = NULL;
        bcompiler_globals->callback = NULL;
        bcompiler_globals->callback_value = NULL;
        bcompiler_globals->callback_key = "NONE";
}

/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(bcompiler)
{
	 
	ZEND_INIT_MODULE_GLOBALS(bcompiler, php_bcompiler_init_globals, NULL);
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(bcompiler)
{
	return SUCCESS;
}
/* }}} */


/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(bcompiler)
{
        return SUCCESS;
}
/* }}} */

/* {{{ PHP_RSHUTDOWN_FUNCTION
 */
PHP_RSHUTDOWN_FUNCTION(bcompiler)
{
	efree(BCOMPILERG(buffer));
        return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(bcompiler)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "bcompiler support", "enabled");
	php_info_print_table_end();

}
/* }}} */




/* ---------------------------------------------------------------------------- 
*
*  MAJOR DEFINES
*
*
* ---------------------------------------------------------------------------- 
*/


int apclib_alignword_int(int x)
{
    typedef union { void* p; int i; } word_t;
    return sizeof(word_t) * (1 + ((x-1)/sizeof(word_t)));
}

/*
These are flags that define start points in the decompile code..
At present theres only 1 - classes ... 
*/
#define BCOMPILER_CLASS_ENTRY  1






/* Definitions that output stuff.  */

/* --------------------------------------- STREAM BASED - eg.  => php 4.3  -------------- */

#ifdef PHP_HAVE_STREAMS


/* SERIALIZE_SCALAR: write a scalar value to dst */
#define SERIALIZE_SCALAR(x, type) {  	\
        if (BCOMPILERG(stream)  != NULL) {             \
                char _buf[255];                                                         \
                 BCOMPILER_DEBUG(("      scalar: %i\n", (int) x));      \
                *(type*) _buf = x;               \
                php_stream_write((php_stream *) BCOMPILERG(stream) ,  _buf, (size_t) apclib_alignword_int(sizeof(type)));   \
        }       \
}

/* DESERIALIZE_SCALAR: read a scalar value from src */
#define DESERIALIZE_SCALAR(xp, type) {                                                          \
        if (php_stream_eof(BCOMPILERG(stream))) {                                           \
                php_error(E_WARNING, "bcompiler: Read Past End of File");               \
                return;                                 \
        }                                       \
        BCOMPILERG(buffer) = (char*) erealloc(BCOMPILERG(buffer), sizeof(type) + 1); \
        php_stream_read((php_stream *) BCOMPILERG(stream) , BCOMPILERG(buffer) , (size_t) apclib_alignword_int(sizeof(type)));   \
	BCOMPILER_DEBUG(("      scalar: %i\n", (int) BCOMPILERG(buffer))); \
        *(xp) = *((type*) BCOMPILERG(buffer));                               \
        BCOMPILER_DEBUG(("     read : %i \n",(type) *(xp)));    \
}


/* STORE_BYTES: memcpy wrapper, writes to dst buffer */
#define STORE_BYTES(bytes, n) {                         \
         if (BCOMPILERG(stream)  != NULL) {     \
                BCOMPILER_DEBUG(("      STORE:  %s \n",bytes));           \
                php_stream_write((php_stream *) BCOMPILERG(stream) , (char *)bytes, n);   \
        }       \
}

	
/* LOAD_BYTES: memcpy wrapper, reads from src buffer */
#define LOAD_BYTES(bytes, n) {                                                          \
        if (php_stream_eof(BCOMPILERG(stream))) {                                           \
                php_error(E_WARNING, "bcompiler: Read Past End of File");               \
                return;                                 \
        }                                                       \
        BCOMPILERG(buffer) = (char*) erealloc(BCOMPILERG(buffer),  n + 1);            \
        php_stream_read((php_stream *) BCOMPILERG(stream) , BCOMPILERG(buffer) , n);   \
        memcpy((char*)bytes, BCOMPILERG(buffer), n);              \
        BCOMPILER_DEBUG(("Got: %s\n",(char*)BCOMPILERG(buffer)));         \
}


#else 








/* Definitions that output stuff.  */

/* --------------------------------------- NON STREAM BASED - eg.  < php 4.3  -------------- */

#define SERIALIZE_SCALAR(x, type) {                                             \
         if (BCOMPILERG(stream)  != NULL) {             \
                char _buf[255];                                                         \
                BCOMPILER_DEBUG(("      scalar: %i\n",(type) x));                                 \
                *(type*) _buf = x;               \
                fwrite(_buf, 1, (size_t) apclib_alignword_int(sizeof(type)), (FILE*) BCOMPILERG(stream));  \
        }       \
}

/* DESERIALIZE_SCALAR: read a scalar value from src */
#define DESERIALIZE_SCALAR(xp, type) {                                                          \
        if (feof((FILE*) BCOMPILERG(stream))) {                                           \
                php_error(E_WARNING, "bcompiler: Read Past End of File");               \
                return;                                 \
        }                                       \
        BCOMPILERG(buffer) = (char*) erealloc(BCOMPILERG(buffer), sizeof(type) + 1);            \
        fread(BCOMPILERG(buffer) , 1, (size_t) apclib_alignword_int(sizeof(type)), (FILE*) BCOMPILERG(stream));    \
        *(xp) = *((type*) BCOMPILERG(buffer));                               \
        BCOMPILER_DEBUG(("     read : %i \n",(type) *(xp)));        \
}


/* STORE_BYTES: memcpy wrapper, writes to dst buffer */
#define STORE_BYTES(bytes, n) {                         \
         if (BCOMPILERG(stream)  != NULL) {     \
                BCOMPILER_DEBUG(("      STORE:  %s \n",bytes));      \
                fwrite((char *)bytes, 1, n, (FILE*) BCOMPILERG(stream));  \
        }       \
}

	
/* LOAD_BYTES: memcpy wrapper, reads from src buffer */
#define LOAD_BYTES(bytes, n) {                                                          \
        if (feof((FILE*) BCOMPILERG(stream))) {                                           \
                php_error(E_WARNING, "bcompiler: Read Past End of File");               \
                return;                                 \
        }                                                       \
        BCOMPILERG(buffer) = (char*) erealloc(BCOMPILERG(buffer),  n + 1);            \
        fread(BCOMPILERG(buffer) ,1, n, (FILE*) BCOMPILERG(stream));  \
        memcpy((char*)bytes, BCOMPILERG(buffer), n);              \
        /* BCOMPILER_DEBUG(("Got: %s\n",(char*)BCOMPILERG(buffer)));   */        \
}


#endif














/* -------------------- Callbacks for bcompiler_parse() ------------------- */

#define CALLBACK_ADD_STRING(str,strlen) {   \
        if (BCOMPILERG(callback) != NULL) {     \
                add_assoc_stringl(BCOMPILERG(callback_value), BCOMPILERG(callback_key), str, strlen , 1);    \
        }       \
}

#define CALLBACK_ADD_KEY_STRING(key, str,strlen) {   \
        if (BCOMPILERG(callback) != NULL) {     \
                add_assoc_stringl(BCOMPILERG(callback_value), key, str, strlen , 1);    \
        }       \
}

#define CALLBACK_ADD_STRING_LONG(str,val) {   \
        char out[255];  \
        int len;     \
        len = sprintf(out,str,val); \
        if (BCOMPILERG(callback) != NULL) {     \
                add_assoc_stringl(BCOMPILERG(callback_value), BCOMPILERG(callback_key),  out, len , 1);    \
        }       \
}
#define CALLBACK_ADD_LONG(key, val) {   \
        if (BCOMPILERG(callback) != NULL) {     \
                add_assoc_long(BCOMPILERG(callback_value), key,  val);    \
        }       \
}
#define CALLBACK_ADD_SUB_LONG(kkey, val) {   \
        char out[255];  \
        strcpy(out, BCOMPILERG(callback_key));  \
        sprintf(out,"%s.%s",out, kkey);   \
        if (BCOMPILERG(callback) != NULL) {     \
                add_assoc_long(BCOMPILERG(callback_value), out, val);    \
        }       \
}
#define CALLBACK_ADD_SUB_DOUBLE(kkey, val) {   \
        char out[255];  \
        strcpy(out, BCOMPILERG(callback_key));  \
        sprintf(out,"%s.%s",out, kkey);   \
        if (BCOMPILERG(callback) != NULL) {     \
                add_assoc_double(BCOMPILERG(callback_value), out, val);    \
        }       \
}
#define CALLBACK_ADD_SUB_STRING(kkey, sval,slen) {   \
        char out[255];  \
        strcpy(out, BCOMPILERG(callback_key));  \
        sprintf(out,"%s.%s",out, kkey);   \
        if (BCOMPILERG(callback) != NULL) {     \
                add_assoc_stringl(BCOMPILERG(callback_value), out, sval ,slen , 1);    \
        }       \
}

 

#define CALLBACK_ADD_DOUBLE(key,val) {   \
        if (BCOMPILERG(callback) != NULL) {     \
                add_assoc_double(BCOMPILERG(callback_value), key,  val);    \
        }       \
}


/* CALLBACK START (opcode) */
#define CALLBACK_START(code) {   \
        if (BCOMPILERG(callback) != NULL) {     \
                        char *str;           \
                        MAKE_STD_ZVAL(BCOMPILERG(callback_value));  \
                        array_init(BCOMPILERG(callback_value));     \
                        /*Optype - */     \
                        str = (char *) getOpcodeName(code );      \
                        add_assoc_long(BCOMPILERG(callback_value), "OPCODE", code);    \
                        add_assoc_stringl(BCOMPILERG(callback_value), "OPCODE_STR", str, strlen(str)  , 1);    \
        }       \
}

#define CALLBACK_END() {   \
        zval myretval;   \
	zval *sendval;	\
        if (BCOMPILERG(callback) != NULL) {     \
		sendval = BCOMPILERG(callback_value);	\
		zval_copy_ctor(sendval);	\
                if(!call_user_function( EG(function_table), NULL, BCOMPILERG(callback), &myretval, 1, &sendval TSRMLS_CC ))  { \
                        zval_dtor(BCOMPILERG(callback_value));      \
                }       \
        }       \
}
 



/* ------------------- All the public php functions ------------------------*/








/* {{{ proto boolean bcompiler_write_class(stream out | filehandle, string class_name)
   Writes compiled data to stream*/
PHP_FUNCTION(bcompiler_write_class)
{
        char *class_name = NULL;
        int class_len;
        zend_class_entry *ce = NULL;
        zval *rsrc;
        
        void *stream;
        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &rsrc, &class_name, &class_len) == FAILURE) {
                return;
        }
#ifdef PHP_HAVE_STREAMS   
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_stream());
#else
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_fopen());
#endif
      
         
        
        zend_str_tolower(class_name, class_len);
        zend_hash_find(EG(class_table), class_name, class_len+1, (void **)&ce);
        if (!ce) {
		php_error(E_WARNING, "Could not find class");
                RETURN_NULL();
        }
        BCOMPILERG(stream)  = stream; 
        BCOMPILERG(callback) = NULL;
        
        SERIALIZE_SCALAR(BCOMPILER_CLASS_ENTRY, char)
        
        apc_serialize_zend_class_entry(ce TSRMLS_CC);
        
        RETURN_TRUE;
}
/* }}} */








/* {{{ proto boolean bcompiler_parse_class(callback, string class_name)
   Writes compiled data to stream*/
PHP_FUNCTION(bcompiler_parse_class)
{
        char *class_name = NULL;
        int class_len;
        zend_class_entry *ce = NULL;
        
        zval *callback = NULL;
        
        
        
        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "zs",   &callback, &class_name, &class_len) == FAILURE) {
                return;
        }
        
        
         
        
        zend_str_tolower(class_name, class_len);
        zend_hash_find(EG(class_table), class_name, class_len+1, (void **)&ce);
        if (!ce) {
                RETURN_NULL();
        }
         
        BCOMPILERG(callback) = callback;
        BCOMPILERG(stream) = NULL;
          
        apc_serialize_zend_class_entry(ce TSRMLS_CC);
        
        RETURN_TRUE;
}
/* }}} */





/* {{{ proto boolean bcompiler_write_header(stream out)
   Writes the start flag*/
PHP_FUNCTION(bcompiler_write_header)
{
         
        
        zval *rsrc;
        
        void *stream;
        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &rsrc) == FAILURE) {
                return;
        }
       
#ifdef PHP_HAVE_STREAMS   
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_stream());
#else
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_fopen());
#endif  

        BCOMPILERG(stream)  = stream; 
        apc_serialize_magic(TSRMLS_C);
        
        RETURN_TRUE;
}
/* }}} */








/* {{{ proto boolean bcompiler_write_end(stream out)
   Writes the end zero*/
PHP_FUNCTION(bcompiler_write_footer)
{
      
        zval *rsrc;
        
        void *stream;
        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &rsrc) == FAILURE) {
                return;
        }
       
#ifdef PHP_HAVE_STREAMS   
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_stream());
#else
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_fopen());
#endif 
        BCOMPILERG(stream)  = stream; 
        
        SERIALIZE_SCALAR(0, char);
        RETURN_TRUE;
}
/* }}} */









/* {{{ proto boolean bcompiler_class(stream out, int start_pos)
   Writes compiled data to stream*/
PHP_FUNCTION(bcompiler_write_exe_footer)
{
        zval *rsrc;
        int start;
        void *stream;
        
        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl", &rsrc, &start) == FAILURE) {
                return;
        }
       
#ifdef PHP_HAVE_STREAMS   
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_stream());
#else
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_fopen());
#endif       
        BCOMPILERG(stream)  = stream; 
        
        SERIALIZE_SCALAR(start, int);
        apc_serialize_magic(TSRMLS_C);
        
        RETURN_TRUE;
}
/* }}} */




/* {{{ proto boolean bcompiler_read(stream in)
   Reads compiled data from stream*/
PHP_FUNCTION(bcompiler_read)
{
        char *class_name = NULL;
        int class_len =0;
        
        zval *rsrc;
        int test =0;
        
        void *stream;
     
        zend_class_entry *ce = NULL;
       
        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &rsrc) == FAILURE) {
                return;
        }
       
#ifdef PHP_HAVE_STREAMS   
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_stream());
#else
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_fopen());
#endif
      
        BCOMPILERG(stream)  = stream; 
        
        
        test = apc_deserialize_magic(TSRMLS_C);
        
        if (test != 0) {
                php_error(E_WARNING, "Could not find Magic header in stream");
                return;
                
        }
        BCOMPILER_DEBUG(("STARTING CLASS\n")); 
        bcompiler_read( TSRMLS_C);     
        
        
        
        RETURN_TRUE;
}
/* }}} */






/* {{{ proto boolean bcompiler_load(bziped file)
   Reads compiled data bziped file*/
PHP_FUNCTION(bcompiler_load)
{
        
        void *stream = NULL;
        int test =0;
#ifdef PHP_HAVE_STREAMS
        zval *bzopen_mode;
#endif      
        zval *rsrc;
        
        zval **args[2]; // used for call array
        zval *zbzopen;  // zval for "bzopen"
        zval *zfilename;
        zval *zmode;
        /* recive args */
        char *filename;
        int filename_len;
        
        

        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",  &filename,&filename_len) == FAILURE) {
                return;
        }
        
        MAKE_STD_ZVAL(zbzopen);
#ifdef PHP_HAVE_STREAMS
        ZVAL_STRING(zbzopen, "bzopen", 1);
#else
        ZVAL_STRING(zbzopen, "fopen", 1);
#endif      
        MAKE_STD_ZVAL(zfilename);
        ZVAL_STRING(zfilename, filename, 1);
        args[0] = &zfilename;
        
        MAKE_STD_ZVAL(zmode);
        ZVAL_STRING(zmode, "r", 1);
        args[1] = &zmode;
        
         
        call_user_function_ex(CG(function_table), NULL,zbzopen,  &rsrc,   2,   args, 0, NULL TSRMLS_CC );
        
#ifdef PHP_HAVE_STREAMS   
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_stream());
#else
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_fopen());
#endif
       
      
        BCOMPILERG(stream)  = stream; 
         
        
        test = apc_deserialize_magic(TSRMLS_C);
        
        if (test != 0) {
                php_error(E_WARNING, "Could not find Magic header in stream");
                return;
                
        }
         
        bcompiler_read( TSRMLS_C);
#ifdef PHP_HAVE_STREAMS        
        php_stream_close(stream); 
#else       
        fclose((FILE *)stream);
#endif  

        RETURN_TRUE;
}
/* }}} */






/* {{{ proto boolean bcompiler_load(bziped file)
   Reads compiled data bziped file*/
PHP_FUNCTION(bcompiler_load_exe)
{
        
        void *stream = NULL;
        int test =0;
#ifdef PHP_HAVE_STREAMS
        zval *bzopen_mode;
#endif        
        zval *rsrc;
        zval *rsrc2;
        
        zval **args[2]; // used for call array

        zval *zbzopen;  // zval for "bzopen"

        zval *zfopen;  // zval for "bzopen"
        zval *zfilename;
        zval *zmode;
        /* recive args */
        char *filename;
        int filename_len;
        
        long pos; //offset to use

        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s",  &filename,&filename_len) == FAILURE) {
                return;
        }
        
        
        
        
        MAKE_STD_ZVAL(zfopen);
        ZVAL_STRING(zfopen, "fopen", 1);
        
        
        MAKE_STD_ZVAL(zfilename);
        ZVAL_STRING(zfilename, filename, 1);
        args[0] = &zfilename;
        
        MAKE_STD_ZVAL(zmode);
        ZVAL_STRING(zmode, "r", 1);
        args[1] = &zmode;
                 
        
        call_user_function_ex(CG(function_table), NULL, zfopen,  &rsrc,   2,   args, 0, NULL TSRMLS_CC );
#ifdef PHP_HAVE_STREAMS   
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_stream());
#else
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc, -1, "File-Handle", php_file_le_fopen());
#endif        
         
      
        BCOMPILERG(stream)  = stream; 
        pos = (long) ((strlen(BCOMPILER_MAGIC_HEADER)+4) * -1); // +4 cause we need to read the string size
        BCOMPILER_DEBUG(("move to start: %d\n", pos));
#ifdef PHP_HAVE_STREAMS         
        php_stream_seek(stream, pos , SEEK_END);
#else       
        fseek((FILE*)stream, pos, SEEK_END);
#endif       
        
        
        test = apc_deserialize_magic(TSRMLS_C);
        
        if (test != 0) {
                php_error(E_WARNING, "Could not find Magic header in stream");
                return;
                
        }
        
        /* now read the length */
        pos = pos - 4; 
        BCOMPILER_DEBUG(("move to start: %d\n", pos));
       
#ifdef PHP_HAVE_STREAMS         
        php_stream_seek(stream, pos , SEEK_END);
#else       
        fseek((FILE*)stream, pos, SEEK_END);
#endif     
        
        
        DESERIALIZE_SCALAR(&pos, int);
        BCOMPILER_DEBUG(("move to start: %d\n", pos)); 
#ifdef PHP_HAVE_STREAMS         
        php_stream_seek(stream, pos , SEEK_SET);
#else       
        fseek((FILE*)stream, pos, SEEK_SET);
#endif  
        
           
        
        
        MAKE_STD_ZVAL(zbzopen);
#ifdef PHP_HAVE_STREAMS
        ZVAL_STRING(zbzopen, "bzopen", 1);
#else       
        ZVAL_STRING(zbzopen, "fopen", 1);
#endif        
        args[0] = &rsrc;
        
        MAKE_STD_ZVAL(zmode);
        ZVAL_STRING(zmode, "r", 1);
        args[1] = &zmode;
        
         
        call_user_function_ex(CG(function_table), NULL,zbzopen,  &rsrc2,   2,   args, 0, NULL TSRMLS_CC );
#ifdef PHP_HAVE_STREAMS   
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc2, -1, "File-Handle", php_file_le_stream());
#else
        ZEND_FETCH_RESOURCE(stream, void *, &rsrc2, -1, "File-Handle", php_file_le_fopen());
#endif          
         
      
        BCOMPILERG(stream)  = stream; 
         
        
        test = apc_deserialize_magic(TSRMLS_C);
        
        if (test != 0) {
                php_error(E_WARNING, "Could not find Magic header in stream");
                return;
                
        }
        bcompiler_read( TSRMLS_C);
       
#ifdef PHP_HAVE_STREAMS        
        php_stream_close(stream); 
#else       
        fclose((FILE *)stream);
#endif   
        RETURN_TRUE;
}
/* }}} */






/* {{{ bcompiler_read - the real reading method.*/


void bcompiler_read(TSRMLS_D) {
        char item;
        zend_class_entry* zc;
        zend_class_entry *ce = NULL;
      
        DESERIALIZE_SCALAR(&item, char);
       
        while (item) {
               
                switch (item) {
                        case BCOMPILER_CLASS_ENTRY:
                                BCOMPILER_DEBUG(("STARTING CLASS\n"));
                        apc_create_zend_class_entry(&zc TSRMLS_CC); 
                        /* is it in the class table already? */
                        
                        zend_hash_find(EG(class_table), zc->name, zc->name_length+1, (void **)&ce);
                        if (ce) {
                                php_error(E_WARNING, "Could not redefine class %s", zc->name);
                                efree(zc);
                        } else {
                                (*zc->refcount)++;
                                BCOMPILER_DEBUG(("ADDING %s \n",zc->name));
                                if (zend_hash_add(EG(class_table), zc->name, zc->name_length+1, zc, sizeof(zend_class_entry), NULL) == FAILURE) { 
                                        (*zc->refcount)--;
                                        zend_hash_destroy(&zc->function_table);
                                        zend_hash_destroy(&zc->default_properties);       
                                        php_error(E_ERROR, "bcompiler: Read Past End of File");
                                }
                        }
        
                }
                DESERIALIZE_SCALAR(&item, char);
        }

}
/* }}} */




/* ---------------------------------------------------------------------------- 
*
*  from apc_lib.h -- GENERALLY STUFF that doesnt need GLOBALS!

*
*
* ---------------------------------------------------------------------------- 
*/



  

char* apclib_estrdup(const char* s)
{
	int len;
	char* dup;

	if (s == NULL) {
		return NULL;
	}
	len = strlen(s);
	dup = (char*) malloc(len+1);
	if (dup == NULL) {
		php_error(E_ERROR, "apclib_estrdup: malloc failed to allocate %u bytes:", len+1);
	}
	memcpy(dup, s, len);
	dup[len] = '\0';
	return dup;
}


void* apclib_erealloc(void* p, size_t n)
{
	p = realloc(p, n);
	if (p == NULL) {
		php_error(E_ERROR, "apclib_erealloc: realloc failed to allocate %u bytes:", n);
	}
	return p;
}

/* type: string (null-terminated) */

void apclib_free_string(char **string) 
{
  if (*string == NULL)
    return;
  efree(*string);
  *string = NULL;
}

 




/*   NO IDEA WHAT THIS STUFF IS ABOUT - probably needs sorting out...

-- it's only used in defered class declarations - which shouldnt really be supported */

/* namearray_t is used to record deferred inheritance relationships */

typedef struct namearray_t namearray_t;

struct namearray_t {
	char* strings;	/* array of null-terminated names */
	int length;		/* logical size of strings */
	int size;		/* physical size of strings */
};

 






/* ---------------------------------------------------------------------------- 
*
*  The name stuff is only really usefull for debugging....
*
*
* ---------------------------------------------------------------------------- 
*/




static const char* opcodes[] = {
        "NOP", /*  0 */
        "ADD", /*  1 */
        "SUB", /*  2 */
        "MUL", /*  3 */
        "DIV", /*  4 */
        "MOD", /*  5 */
        "SL", /*  6 */
        "SR", /*  7 */
        "CONCAT", /*  8 */
        "BW_OR", /*  9 */
        "BW_AND", /*  10 */
        "BW_XOR", /*  11 */
        "BW_NOT", /*  12 */
        "BOOL_NOT", /*  13 */
        "BOOL_XOR", /*  14 */
        "IS_IDENTICAL", /*  15 */
        "IS_NOT_IDENTICAL", /*  16 */
        "IS_EQUAL", /*  17 */
        "IS_NOT_EQUAL", /*  18 */
        "IS_SMALLER", /*  19 */
        "IS_SMALLER_OR_EQUAL", /*  20 */
        "CAST", /*  21 */
        "QM_ASSIGN", /*  22 */
        "ASSIGN_ADD", /*  23 */
        "ASSIGN_SUB", /*  24 */
        "ASSIGN_MUL", /*  25 */
        "ASSIGN_DIV", /*  26 */
        "ASSIGN_MOD", /*  27 */
        "ASSIGN_SL", /*  28 */
        "ASSIGN_SR", /*  29 */
        "ASSIGN_CONCAT", /*  30 */
        "ASSIGN_BW_OR", /*  31 */
        "ASSIGN_BW_AND", /*  32 */
        "ASSIGN_BW_XOR", /*  33 */
        "PRE_INC", /*  34 */
        "PRE_DEC", /*  35 */
        "POST_INC", /*  36 */
        "POST_DEC", /*  37 */
        "ASSIGN", /*  38 */
        "ASSIGN_REF", /*  39 */
        "ECHO", /*  40 */
        "PRINT", /*  41 */
        "JMP", /*  42 */
        "JMPZ", /*  43 */
        "JMPNZ", /*  44 */
        "JMPZNZ", /*  45 */
        "JMPZ_EX", /*  46 */
        "JMPNZ_EX", /*  47 */
        "CASE", /*  48 */
        "SWITCH_FREE", /*  49 */
        "BRK", /*  50 */
        "CONT", /*  51 */
        "BOOL", /*  52 */
        "INIT_STRING", /*  53 */
        "ADD_CHAR", /*  54 */
        "ADD_STRING", /*  55 */
        "ADD_VAR", /*  56 */
        "BEGIN_SILENCE", /*  57 */
        "END_SILENCE", /*  58 */
        "INIT_FCALL_BY_NAME", /*  59 */
        "DO_FCALL", /*  60 */
        "DO_FCALL_BY_NAME", /*  61 */
        "RETURN", /*  62 */
        "RECV", /*  63 */
        "RECV_INIT", /*  64 */
        "SEND_VAL", /*  65 */
        "SEND_VAR", /*  66 */
        "SEND_REF", /*  67 */
        "NEW", /*  68 */
        "JMP_NO_CTOR", /*  69 */
        "FREE", /*  70 */
        "INIT_ARRAY", /*  71 */
        "ADD_ARRAY_ELEMENT", /*  72 */
        "INCLUDE_OR_EVAL", /*  73 */
        "UNSET_VAR", /*  74 */
        "UNSET_DIM_OBJ", /*  75 */
        "ISSET_ISEMPTY", /*  76 */
        "FE_RESET", /*  77 */
        "FE_FETCH", /*  78 */
        "EXIT", /*  79 */
        "FETCH_R", /*  80 */
        "FETCH_DIM_R", /*  81 */
        "FETCH_OBJ_R", /*  82 */
        "FETCH_W", /*  83 */
        "FETCH_DIM_W", /*  84 */
        "FETCH_OBJ_W", /*  85 */
        "FETCH_RW", /*  86 */
        "FETCH_DIM_RW", /*  87 */
        "FETCH_OBJ_RW", /*  88 */
        "FETCH_IS", /*  89 */
        "FETCH_DIM_IS", /*  90 */
        "FETCH_OBJ_IS", /*  91 */
        "FETCH_FUNC_ARG", /*  92 */
        "FETCH_DIM_FUNC_ARG", /*  93 */
        "FETCH_OBJ_FUNC_ARG", /*  94 */
        "FETCH_UNSET", /*  95 */
        "FETCH_DIM_UNSET", /*  96 */
        "FETCH_OBJ_UNSET", /*  97 */
        "FETCH_DIM_TMP_VAR", /*  98 */
        "FETCH_CONSTANT", /*  99 */
        "DECLARE_FUNCTION_OR_CLASS", /*  100 */
        "EXT_STMT", /*  101 */
        "EXT_FCALL_BEGIN", /*  102 */
        "EXT_FCALL_END", /*  103 */
        "EXT_NOP", /*  104 */
        "TICKS", /*  105 */
        "SEND_VAR_NO_REF", /*  106 */
                  
        "ZEND_CATCH", /*                        107 */
        "ZEND_THROW", /*                        108 */
        "ZEND_FETCH_CLASS", /*                  109 */
        "ZEND_CLONE", /*                        110 */
        "ZEND_INIT_CTOR_CALL", /*               111 */
        "ZEND_INIT_METHOD_CALL", /*             112 */
        "ZEND_INIT_STATIC_METHOD_CALL", /*      113 */
        "ZEND_ISSET_ISEMPTY_VAR", /*            114 */
        "ZEND_ISSET_ISEMPTY_DIM_OBJ", /*        115 */
        "ZEND_IMPORT_FUNCTION", /*              116 */
        "ZEND_IMPORT_CLASS", /*                 117 */
        "ZEND_IMPORT_CONST", /*                 118 */
        
        "**UNKNOWN**", /*                 119 */
        "**UNKNOWN**", /*                 120 */
        
        "ZEND_ASSIGN_ADD_OBJ", /*               121 */
        "ZEND_ASSIGN_SUB_OBJ", /*               122 */
        "ZEND_ASSIGN_MUL_OBJ", /*               123 */
        "ZEND_ASSIGN_DIV_OBJ", /*               124 */
        "ZEND_ASSIGN_MOD_OBJ", /*               125 */
        "ZEND_ASSIGN_SL_OBJ", /*                126 */
        "ZEND_ASSIGN_SR_OBJ", /*                127 */
        "ZEND_ASSIGN_CONCAT_OBJ", /*            128 */
        "ZEND_ASSIGN_BW_OR_OBJ", /*             129 */
        "END_ASSIGN_BW_AND_OBJ", /*             130 */
        "END_ASSIGN_BW_XOR_OBJ", /*             131 */
        "ZEND_PRE_INC_OBJ", /*                  132 */
        "ZEND_PRE_DEC_OBJ", /*                  133 */
        "ZEND_POST_INC_OBJ", /*                 134 */
        "ZEND_POST_DEC_OBJ", /*                 135 */
        "ZEND_ASSIGN_OBJ", /*                   136 */
        "ZEND_MAKE_VAR", /*                     137 */
        
                 
          
          
        "BEGIN_HASH_DEF", /*  138 */
        "OPCODE_ARRAY"  /*  139 */
  
};
 

#define  ZEND_BEGIN_HASH_DEF 138
#define  ZEND_OPCODE_ARRAY 139
static const int NUM_OPCODES = sizeof(opcodes) / sizeof(opcodes[0]);





const char* getOpcodeName(int op)
{
  if (op < 0 || op >= NUM_OPCODES) {
    return "(unknown)";
  }
  return opcodes[op];
}






 /* inherit: recursively inherit methods from base to all the children
 * of parent, the children of parent's children, and so on */
 
static void bcompiler_inherit(zend_class_entry* base  TSRMLS_DC)
{
 	
	 
	zend_class_entry* child;
	BCOMPILER_DEBUG(("base: %s",base->name));
         
	if (!base->parent) {
		return;		// 'parent' has no children, nothing to do 
	}
        child = base->parent;
        
        while(child) {
                BCOMPILER_DEBUG(("child: %s",child->name));
		zend_do_inheritance(base,child);
                child =  child->parent;
	}
}

 
 	









/* ---------------------------------------------------------------------------- 
*
*  START STUFF THAT NEEDS GLOBALS.... (eg. TSRMLS_DC)
*
*
* ---------------------------------------------------------------------------- 
*/


   

void apc_serialize_string(char* string TSRMLS_DC)
{
	int len;

	/* by convention, mark null strings with a length of -1 */
	if (string == NULL) {
		SERIALIZE_SCALAR(-1, int);
		return;
	}

	len = strlen(string);
        if (len <= 0) {
                SERIALIZE_SCALAR(-1, int);
                return;
        }
        BCOMPILER_DEBUG(("STRING: len, string : %s\n", string)); 
	SERIALIZE_SCALAR(len, int);
	STORE_BYTES(string, len);
}

void apc_serialize_zstring(char* string, int len TSRMLS_DC)
{
  /* by convention, mark null strings with a length of -1 */
  if (len <= 0 || string == NULL) {
    SERIALIZE_SCALAR(-1, int);
    return;
  }
//  len = strlen(string);
  SERIALIZE_SCALAR(len, int);
  STORE_BYTES(string, len);
}

void apc_create_string(char** string TSRMLS_DC)
{
  int len = 0;

  DESERIALIZE_SCALAR(&len, int);
  if (len <= 0) {
    /* by convention, empty marked strings (len is -1)  are returned as an empty string */
    *string = (char*) emalloc(1);
    (*string)[0] = '\0';
    return;
  }
  *string = (char*) emalloc(len + 1);
  LOAD_BYTES(*string, len);
  (*string)[len] = '\0';
}

void apc_serialize_arg_types(zend_uchar* arg_types TSRMLS_DC)
{
	if (arg_types == NULL) {
		SERIALIZE_SCALAR(0, char);
		return; /* arg_types is null */
	}
	SERIALIZE_SCALAR(1, char);
	SERIALIZE_SCALAR(arg_types[0], zend_uchar);
	STORE_BYTES(arg_types + 1, arg_types[0]*sizeof(zend_uchar));
}

void apc_create_arg_types(zend_uchar** arg_types TSRMLS_DC)
{
	char exists;
	zend_uchar count;

	DESERIALIZE_SCALAR(&exists, char);
	if (!exists) {
		*arg_types = NULL;
		return; /* arg_types is null */
	}
	DESERIALIZE_SCALAR(&count, zend_uchar);
	*arg_types = emalloc(count + 1);
	(*arg_types)[0] = count;
	LOAD_BYTES((*arg_types) + 1, count*sizeof(zend_uchar));
}

/* precompiler routines */
void apc_serialize_magic(TSRMLS_D)
{
	apc_serialize_string(BCOMPILER_MAGIC_HEADER TSRMLS_CC);
}

int apc_deserialize_magic(TSRMLS_D)
{
	char *tmp;
	int retval;

	apc_create_string(&tmp TSRMLS_CC);
        BCOMPILER_DEBUG(("got %s\n", tmp));
	retval = strcmp(tmp,BCOMPILER_MAGIC_HEADER);
	apclib_free_string(&tmp); // MPB
	return retval;
}
/* type: zend_llist */

static void store_zend_llist_element(void* arg, void* data  TSRMLS_DC)
{
	int size = *((int*)arg);
	STORE_BYTES((char*)data, size);
}


void apc_serialize_zend_llist(zend_llist* list TSRMLS_DC)
{
	char exists;

	exists = (list != NULL) ? 1 : 0;
        SERIALIZE_SCALAR(exists, char);  // write twice to remove need for peeking.
	
	if (!exists) {
		return;
	}
        SERIALIZE_SCALAR(exists, char);
	SERIALIZE_SCALAR(list->size, size_t);
	SERIALIZE_SCALAR(list->dtor, void*);
	SERIALIZE_SCALAR(list->persistent, unsigned char);
	SERIALIZE_SCALAR(zend_llist_count(list), int);
	zend_llist_apply_with_argument(list, store_zend_llist_element,
		&list->size  TSRMLS_CC);
}

void apc_deserialize_zend_llist(zend_llist* list TSRMLS_DC)
{
	char exists;
	size_t size;
	void (*dtor)(void*);
	unsigned char persistent;
	int count, i;
	char* data;

	DESERIALIZE_SCALAR(&exists, char);
	assert(exists != 0); 

	/* read the list parameters */
	DESERIALIZE_SCALAR(&size, size_t);
	DESERIALIZE_SCALAR(&dtor, void*);
	DESERIALIZE_SCALAR(&persistent, unsigned char);
	/* initialize the list */
	zend_llist_init(list, size, dtor, persistent);

	/* insert the list elements */
	DESERIALIZE_SCALAR(&count, int);
	data = (char*) malloc(list->size);
	for (i = 0; i < count; i++) {
		LOAD_BYTES(data, list->size);
		zend_llist_add_element(list, data);
	}
	free(data);
}

void apc_create_zend_llist(zend_llist** list TSRMLS_DC)
{
	char exists;

	/* Sneak a look one byte ahead to see whether the list exists or not.
	 * If it does, then exists is part of the structure, otherwise it was
	 * a zero-value placeholder byte. */
	/*PEEK_SCALAR(&exists, char);  */
        DESERIALIZE_SCALAR(&exists, char); 
	if (exists) {
		*list = (zend_llist*) emalloc(sizeof(zend_llist));
		apc_deserialize_zend_llist(*list TSRMLS_CC);
	}
	else {
		/* DESERIALIZE_SCALAR(&exists, char);  */
		*list = 0;
	}
}


/* type: HashTable */

 

/* type: HashTable */

 


 






	
void apc_serialize_hashtable(HashTable* ht, void* funcptr TSRMLS_DC)
{
	char exists;	/* for identifying null lists */
	Bucket* p;
	void (*serialize_bucket)(void* TSRMLS_DC);
        

     
	serialize_bucket = (void(*)(void* TSRMLS_DC)) funcptr;

	exists = (ht != NULL) ? 1 : 0;
	SERIALIZE_SCALAR(exists, char);
	if (!exists) {
		return;
	}
         
        
       
	/* Serialize the hash meta-data. */
	SERIALIZE_SCALAR(ht->nTableSize, uint);
	SERIALIZE_SCALAR(ht->pDestructor, void*);
	SERIALIZE_SCALAR(ht->nNumOfElements, uint);
	SERIALIZE_SCALAR(ht->persistent, int);

	/* Iterate through the buckets of the hash, serializing as we go. */
	p = ht->pListHead;
	while(p != NULL) {
		SERIALIZE_SCALAR(p->h, ulong);
		SERIALIZE_SCALAR(p->nKeyLength,uint);
                CALLBACK_START(ZEND_BEGIN_HASH_DEF);
		CALLBACK_ADD_LONG("LINE", -1);
                BCOMPILERG(callback_key) = "HASH";
                CALLBACK_ADD_STRING(p->arKey, p->nKeyLength);
                CALLBACK_END();
		apc_serialize_zstring(p->arKey, p->nKeyLength TSRMLS_CC);
		serialize_bucket(p->pData TSRMLS_CC ); 
		p = p->pListNext;
	}
        
         
        
        
}

void apc_deserialize_hashtable(HashTable* ht, void* funcptr, void* dptr, int datasize, char exists TSRMLS_DC)
{
	 
	uint nSize;
	dtor_func_t pDestructor;
	uint nNumOfElements;
	int persistent;
	int j;
	ulong h;
	uint nKeyLength;
	char* arKey;
	void* pData;
	int status;
	void (*deserialize_bucket)(void*);
	void (*free_bucket)(void*);
        BCOMPILER_DEBUG(("-----------------------------\nHASH TABLE:\n"));
	deserialize_bucket = (void(*)(void*)) funcptr;
	free_bucket = (void(*)(void*)) dptr;
        
	 
	assert(exists != 0);

	DESERIALIZE_SCALAR(&nSize, uint);
	DESERIALIZE_SCALAR(&pDestructor, void*);
	DESERIALIZE_SCALAR(&nNumOfElements,uint);
	DESERIALIZE_SCALAR(&persistent, int);
	
        BCOMPILER_DEBUG(("-----------------------------\nBEGIN TABLE:\n"));
        
	/* Although the hash is already allocated (we're a deserialize, not a 
	 * create), we still need to initialize it. If this fails, something 
	 * very very bad happened. */
	status = zend_hash_init(ht, nSize, NULL, pDestructor, persistent);
	assert(status != FAILURE);
	
	/* Luckily, the number of elements in a hash is part of its struct, so
	 * we can just deserialize that many hashtable elements. */

	for (j = 0; j < (int) nNumOfElements; j++) {
		DESERIALIZE_SCALAR(&h, ulong);
		DESERIALIZE_SCALAR(&nKeyLength, uint);
		apc_create_string(&arKey TSRMLS_CC);
		deserialize_bucket(&pData);

		/* If nKeyLength is non-zero, this element is a hashed key/value
		 * pair. Otherwise, it is an array element with a numeric index */
                BCOMPILER_DEBUG(("-----------------------------\nFUNC %s\n",arKey));
	
		if (nKeyLength != 0) {
			if(datasize == sizeof(void*)) {
				zend_hash_add_or_update(ht, arKey, nKeyLength, &pData,
				                        datasize, NULL, HASH_ADD);
			}
			else {
				zend_hash_add_or_update(ht, arKey, nKeyLength, pData,
				                        datasize, NULL, HASH_ADD);
			}
		}
		else {	/* numeric index, not key string */
			if(datasize == sizeof(void*)) {
				zend_hash_index_update(ht, h, &pData, datasize, NULL);
			}
			else {
				zend_hash_index_update(ht, h, pData, datasize, NULL);
			}
		}
                if (dptr != NULL) {
                  free_bucket(&pData); 
                }
                // else {
                //   BCOMPILER_DEBUG((stderr, "Do not free: %s\n", arKey));
                // }
                apclib_free_string(&arKey); // ADDED BY MPB
	}
}

void apc_create_hashtable(HashTable** ht, void* funcptr, void* dptr, int datasize TSRMLS_DC)
{
	char exists;	/* for identifying null hashtables */

	/*PEEK_SCALAR(&exists, char); */
	DESERIALIZE_SCALAR(&exists, char);
        if (exists==1) {
		*ht = (HashTable*) emalloc(sizeof(HashTable));
		apc_deserialize_hashtable(*ht, funcptr, dptr, datasize,exists TSRMLS_CC);
	}
	else {
                /* DESERIALIZE_SCALAR(&exists, char); */
		*ht = 0;
	}
}

/* type: zvalue_value */


void apc_serialize_zvalue_value(zvalue_value* zv, int type TSRMLS_DC)
{
	/* A zvalue_value is a union, and as such we first need to
	 * determine exactly what it's type is, then serialize the
	 * appropriate structure. */
        CALLBACK_ADD_SUB_LONG("u.constant.type",type);
	switch (type) {
	  case IS_RESOURCE:
	  case IS_BOOL:
	  case IS_LONG:
                CALLBACK_ADD_SUB_LONG("u.constant.val",zv->lval);
                CALLBACK_ADD_LONG(BCOMPILERG(callback_key),zv->lval);
		SERIALIZE_SCALAR(zv->lval, long);
		break;	
	  case IS_DOUBLE:
                CALLBACK_ADD_SUB_DOUBLE("u.constant.val",zv->dval);
                CALLBACK_ADD_DOUBLE(BCOMPILERG(callback_key),zv->dval);
		SERIALIZE_SCALAR(zv->dval, double);
		break;
	  case IS_NULL:
		/* null value, do nothing */
		break;
	  case IS_CONSTANT:
	  case IS_STRING:
	  case FLAG_IS_BC:
                CALLBACK_ADD_SUB_STRING("u.constant.val",zv->str.val, zv->str.len);
                CALLBACK_ADD_STRING(zv->str.val, zv->str.len);
		apc_serialize_zstring(zv->str.val, zv->str.len  TSRMLS_CC);
		SERIALIZE_SCALAR(zv->str.len, int);
		break;
	  case IS_ARRAY:
                CALLBACK_ADD_STRING("ARRAY?", 6);
		apc_serialize_hashtable(zv->ht, apc_serialize_zval_ptr  TSRMLS_CC);
		break;
	  case IS_CONSTANT_ARRAY:
                CALLBACK_ADD_STRING("CARRAY?", 7);
		apc_serialize_hashtable(zv->ht, apc_serialize_zval_ptr TSRMLS_CC);
		break;
	  case IS_OBJECT:
                CALLBACK_ADD_STRING("OBJECT?", 7);
		apc_serialize_zend_class_entry(zv->obj.ce TSRMLS_CC);
		apc_serialize_hashtable(zv->obj.properties, apc_serialize_zval_ptr TSRMLS_CC);
		break;
	  default:
		/* The above list enumerates all types.  If we get here,
		 * something very very bad has happened. */
		assert(0);
	}
}

void apc_deserialize_zvalue_value(zvalue_value* zv, int type TSRMLS_DC)
{
	/* We peeked ahead in the calling routine to deserialize the
	 * type. Now we just deserialize. */
	switch(type) {
	  case IS_RESOURCE:
	  case IS_BOOL:
	  case IS_LONG:
		DESERIALIZE_SCALAR(&zv->lval, int);
		break;
	  case IS_NULL:
		/* null value, do nothing */
		break;
	  case IS_DOUBLE:
		DESERIALIZE_SCALAR(&zv->dval, double);
		break;
	  case IS_CONSTANT:
	  case IS_STRING:
	  case FLAG_IS_BC:
		apc_create_string(&zv->str.val TSRMLS_CC);
		DESERIALIZE_SCALAR(&zv->str.len, int);
		break;
	  case IS_ARRAY:
		apc_create_hashtable(&zv->ht, apc_create_zval, NULL, sizeof(void*) TSRMLS_CC);
		break;
	  case IS_CONSTANT_ARRAY:
		apc_create_hashtable(&zv->ht, apc_create_zval, NULL, sizeof(void*) TSRMLS_CC);
		break;
	  case IS_OBJECT:
		apc_create_zend_class_entry(&zv->obj.ce TSRMLS_CC);
		apc_create_hashtable(&zv->obj.properties, apc_create_zval, NULL, sizeof(zval *) TSRMLS_CC);
		break;
	  default:
        /* The above list enumerates all types.  If we get here,
         * something very very bad has happened. */
		assert(0);
	}	
}


/* type: zval */

void apc_serialize_zval_ptr(zval** zv TSRMLS_DC)
{
	apc_serialize_zval(*zv TSRMLS_CC);
}


	
void apc_serialize_zval(zval* zv TSRMLS_DC)
{
	/* type is the switch for serializing zvalue_value */
	SERIALIZE_SCALAR(zv->type, zend_uchar);
	apc_serialize_zvalue_value(&zv->value, zv->type TSRMLS_CC);
	SERIALIZE_SCALAR(zv->is_ref, zend_uchar);
	SERIALIZE_SCALAR(zv->refcount, zend_ushort);
}

void apc_deserialize_zval(zval* zv TSRMLS_DC)
{
	/* type is the switch for deserializing zvalue_value */
	DESERIALIZE_SCALAR(&zv->type, zend_uchar);
	apc_deserialize_zvalue_value(&zv->value, zv->type TSRMLS_CC);
	DESERIALIZE_SCALAR(&zv->is_ref, zend_uchar);
	DESERIALIZE_SCALAR(&zv->refcount, zend_ushort);
}

void apc_create_zval(zval** zv TSRMLS_DC)
{
	*zv = (zval*) emalloc(sizeof(zval));
	apc_deserialize_zval(*zv TSRMLS_CC);
}
void apc_free_zval(zval** zv TSRMLS_DC)
{
  if (*zv != NULL) {
    switch((*zv)->type) {
      case IS_CONSTANT:
      case IS_STRING:
      case FLAG_IS_BC:
        apclib_free_string(&(*zv)->value.str.val);
        break;
    }
    efree(*zv);
  }
  *zv = NULL;
}



/* type: zend_function_entry */


void apc_serialize_zend_function_entry(zend_function_entry* zfe TSRMLS_DC)
{
	apc_serialize_string(zfe->fname TSRMLS_CC);
        BCOMPILER_DEBUG(("serializing function %s\n",zfe->fname ));
	SERIALIZE_SCALAR(zfe->handler, void*);
	apc_serialize_arg_types(zfe->func_arg_types TSRMLS_CC);
}

void apc_deserialize_zend_function_entry(zend_function_entry* zfe TSRMLS_DC)
{
	apc_create_string(&zfe->fname TSRMLS_CC);
	DESERIALIZE_SCALAR(&zfe->handler, void*);
        BCOMPILER_DEBUG(("deserializing function %s\n",zfe->handler));
	apc_create_arg_types(&zfe->func_arg_types TSRMLS_CC);
}


/* type: zend_property_reference */


void apc_serialize_zend_property_reference(zend_property_reference* zpr TSRMLS_DC)
{
	SERIALIZE_SCALAR(zpr->type, int);
	apc_serialize_zval(zpr->object TSRMLS_CC);
	apc_serialize_zend_llist(zpr->elements_list TSRMLS_CC);
}

void apc_deserialize_zend_property_reference(zend_property_reference* zpr TSRMLS_DC)
{
	SERIALIZE_SCALAR(zpr->type, int);
	apc_deserialize_zval(zpr->object TSRMLS_CC);
	apc_create_zend_llist(&zpr->elements_list TSRMLS_CC);
}


/* type: zend_overloaded_element - i dont think these are used!*/


void apc_serialize_zend_overloaded_element(zend_overloaded_element* zoe TSRMLS_DC)
{
	SERIALIZE_SCALAR(zoe->type, zend_uchar);
	apc_serialize_zval(&zoe->element TSRMLS_CC);
}

void apc_deserialize_zend_overloaded_element(zend_overloaded_element* zoe TSRMLS_DC)
{
	DESERIALIZE_SCALAR(&zoe->type, zend_uchar);
	apc_deserialize_zval(&zoe->element TSRMLS_CC);
}


/* type: zend_class_entry */

	
void apc_serialize_zend_class_entry(zend_class_entry* zce TSRMLS_DC)
{
        /* other than sending to a stream it also does a callback! 
        
        The Array for a Class entry should look like:
         NOT::Type? - what is this used for?
         Name
         Parent
         NOT::refcount[0] ??
         NOT::constants_updated ??
         NOT::call,set etc: - would these point to a 'method?' !!
        ZEND_DECLARE_FUNCTION_OR_CLASS
        */
      
       
        
        zend_function_entry* zfe;
        int count, i, exists;
         
        
        
        
         
        BCOMPILER_DEBUG(("adding type : %i\n",zce->type )); 
        SERIALIZE_SCALAR(zce->type, char);
        apc_serialize_zstring(zce->name, zce->name_length TSRMLS_CC);
        SERIALIZE_SCALAR(zce->name_length, uint);
        
        /* Serialize the name of this class's parent class (if it has one)
         * so that we can perform inheritance during deserialization (see
         * apc_deserialize_zend_class_entry). */
        
        exists = (zce->parent != NULL) ? 1 : 0;
        SERIALIZE_SCALAR(exists, char);
        
        if (exists) {
                /*Parent*/ 
               
                apc_serialize_zstring(zce->parent->name, zce->parent->name_length TSRMLS_CC);
        }
        
        CALLBACK_START(zce->type);   
        BCOMPILERG(callback_key) = "CLASS";
        CALLBACK_ADD_STRING(zce->name, zce->name_length);         
        CALLBACK_END();   
        /* now callback to class/funct etc. */
         
        
        
        
        SERIALIZE_SCALAR(zce->refcount[0], int);
        SERIALIZE_SCALAR(zce->constants_updated, zend_bool);
        BCOMPILER_DEBUG(("-----------------------------\nFUNC TABLE:\n")); 
        apc_serialize_hashtable(&zce->function_table, apc_serialize_zend_function TSRMLS_CC);
        BCOMPILER_DEBUG(("-----------------------------\nVARS:\n"));
        apc_serialize_hashtable(&zce->default_properties, apc_serialize_zval_ptr TSRMLS_CC);

	/* zend_class_entry.builtin_functions: this array appears to be
	 * terminated by an element where zend_function_entry.fname is null 
	 * First we count the number of elements, then we serialize that count
	 * and all the functions. */
        BCOMPILER_DEBUG(("-----------------------------\nBUILTIN:\n"));
	count = 0;
	if (zce->builtin_functions) {
		for (zfe = zce->builtin_functions; zfe->fname != NULL; zfe++) {
			count++;
		}
	}
        
	SERIALIZE_SCALAR(count, int);
	for (i = 0; i < count; i++) {
		apc_serialize_zend_function_entry(&zce->builtin_functions[i] TSRMLS_CC);
	}
        
        SERIALIZE_SCALAR(zce->handle_function_call, void*);
        SERIALIZE_SCALAR(zce->handle_property_get, void*);
        SERIALIZE_SCALAR(zce->handle_property_set, void*);
}

void apc_deserialize_zend_class_entry(zend_class_entry* zce TSRMLS_DC)
{
        
        int count, i, exists;

        DESERIALIZE_SCALAR(&zce->type, char);
        apc_create_string(&zce->name TSRMLS_CC);
        DESERIALIZE_SCALAR(&zce->name_length, uint);

        DESERIALIZE_SCALAR(&exists, char);
	if (exists) {
		
		// This class knows the name of its parent, most likely because
		// its parent is defined in the same source file. Therefore, we
		// can simply restore the parent link, and not worry about
		// manually inheriting methods from the parent (PHP will perform
		// the inheritance). 
		 	 
		char* parent_name;	// name of parent class 

		apc_create_string(&parent_name TSRMLS_CC);
		
		/*if ((zend_hash_find(CG(class_table), parent_name,
			strlen(parent_name) + 1, (void**) &zce->parent) == FAILURE)) {
                        php_error(E_ERROR, "unable to inherit %s", parent_name);
                }*/
		apclib_free_string(&parent_name); 
		zce->parent = NULL;
        } else {
                zce->parent = NULL;
        }

        /* refcount is a pointer to a single int.  Don't ask me why, I
        // just work here;. */
 
        zce->refcount = (int*) emalloc(sizeof(int));
        DESERIALIZE_SCALAR(zce->refcount, int);
        DESERIALIZE_SCALAR(&zce->constants_updated, zend_bool);
         BCOMPILER_DEBUG(("-----------------------------\nFUNC TABLE:\n"));
         DESERIALIZE_SCALAR(&exists, char);
	
	apc_deserialize_hashtable(&zce->function_table, (void*) apc_create_zend_function, (void*) apc_free_zend_function,
		(int) sizeof(zend_function) , (char) exists TSRMLS_CC);
        BCOMPILER_DEBUG(("-----------------------------\nDEFAYLT PROPS:\n")); 
	 DESERIALIZE_SCALAR(&exists, char);
	apc_deserialize_hashtable(&zce->default_properties,  (void*) apc_create_zval,  (void*) NULL, (int) sizeof(zval *),  (char)exists TSRMLS_CC);

	// see apc_serialize_zend_class_entry() for a description of the
	// zend_class_entry.builtin_functions member 

	DESERIALIZE_SCALAR(&count, int);
	zce->builtin_functions = NULL;
	if (count > 0) {
		zce->builtin_functions = (zend_function_entry*)
			emalloc((count+1) * sizeof(zend_function_entry));
		zce->builtin_functions[count].fname = NULL;
		for (i = 0; i < count; i++) {
			apc_deserialize_zend_function_entry(&zce->builtin_functions[i] TSRMLS_CC);
		}
	}
	
	DESERIALIZE_SCALAR(&zce->handle_function_call, void*);
	DESERIALIZE_SCALAR(&zce->handle_property_get, void*);
	DESERIALIZE_SCALAR(&zce->handle_property_set, void*);

  
	/* Resolve the inheritance relationships that have been delayed and
	 * are waiting for this class to be created, i.e., every child of
	 * this class that has already been compiled needs to be inherited
	 * from this class. Call inherit() with this class as the base class
	 * (first parameter) and as the current parent class (second parameter).
	 * inherit will recursively resolve every inheritance relationships
	 * in which this class is the base. */

	  /*
	  Note this will never happen, as compiled classes include all the functions, vars
	  etc. from their inherited children anyway.
	  */
          
        if (zce->parent) {
                bcompiler_inherit(zce TSRMLS_CC); 
        }
}

void apc_create_zend_class_entry(zend_class_entry** zce TSRMLS_DC)
{
	  
	*zce = (zend_class_entry*) emalloc(sizeof(zend_class_entry));
	apc_deserialize_zend_class_entry(*zce TSRMLS_CC);
	
}

 
  

/* type: znode */


void apc_serialize_znode(znode* zn TSRMLS_DC)
{
	SERIALIZE_SCALAR(zn->op_type, int);

	/* If the znode's op_type is IS_CONST, we know precisely what it is.
	 * otherwise, it is too case-dependent (or inscrutable), so we do
	 * a bitwise copy. */
        CALLBACK_ADD_SUB_LONG("op_type" ,(long) zn->op_type);  
        CALLBACK_ADD_SUB_LONG("u.fetch_type" ,(int)  zn->u.fetch_type);  
         CALLBACK_ADD_SUB_LONG("u.EA.type" ,(int)  zn->u.EA.type);  
                     
	switch(zn->op_type) {
                case IS_CONST: 
                        if (&zn->u.fetch_type == ZEND_FETCH_GLOBAL) {
                                CALLBACK_ADD_STRING("**GLOBAL",8);  
                        }
                        apc_serialize_zval(&zn->u.constant TSRMLS_CC);
                        break;
                        
                case IS_VAR:
                        CALLBACK_ADD_SUB_LONG("u.var" ,(long) zn->u.var);  
                        CALLBACK_ADD_STRING_LONG("VAR:%i",(long) zn->u.var); 
                        STORE_BYTES(&zn->u, sizeof(zn->u));
                        break;
                case IS_TMP_VAR:
                        CALLBACK_ADD_SUB_LONG("u.var" ,(long) zn->u.var);  
                        CALLBACK_ADD_STRING_LONG("TMPVAR:%i",(long) zn->u.var); 
                        STORE_BYTES(&zn->u, sizeof(zn->u));
                        break; 
                case IS_UNUSED:
                        /* CALLBACK_ADD_STRING("**NOTUSED",9);  */
                        CALLBACK_ADD_SUB_LONG("u.opline_num" ,(long) zn->u.opline_num);  
                        
                        STORE_BYTES(&zn->u, sizeof(zn->u));
                        break;
                        
                default:
                        CALLBACK_ADD_STRING_LONG("**UNKNOWN",(long) sizeof(zn->u)); 
                        STORE_BYTES(&zn->u, sizeof(zn->u));
                        break;
	}
}

void apc_deserialize_znode(znode* zn TSRMLS_DC)
{
	DESERIALIZE_SCALAR(&zn->op_type, int);

	/* If the znode's op_type is IS_CONST, we know precisely what it is.
	 * otherwise, it is too case-dependent (or inscrutable), so we do
	 * a bitwise copy. */
	
	switch(zn->op_type) {
	  case IS_CONST:
		apc_deserialize_zval(&zn->u.constant TSRMLS_CC);
		break;
	  default:
		LOAD_BYTES(&zn->u, sizeof(zn->u));
		break;
	}
}


/* type: zend_op */


void apc_serialize_zend_op(int id, zend_op* zo TSRMLS_DC)
{
         
	//fprintf(stderr, "Outputting %s\n", getOpcodeName(zo->opcode));
        CALLBACK_START(zo->opcode);
        SERIALIZE_SCALAR(zo->opcode, zend_uchar);
        
        BCOMPILERG(callback_key)="RESULT";
	 apc_serialize_znode(&zo->result TSRMLS_CC);
        
        BCOMPILERG(callback_key)="OP1";
  	 apc_serialize_znode(&zo->op1 TSRMLS_CC);
        
        BCOMPILERG(callback_key)="OP2";
	 apc_serialize_znode(&zo->op2 TSRMLS_CC);
        
	 SERIALIZE_SCALAR(zo->extended_value, ulong);
        if (zo->extended_value != 0) {
                CALLBACK_ADD_LONG("EXTENDED",(long) zo->extended_value);  
        }
        CALLBACK_ADD_LONG("LINE",(long) id);  
        /* add some useful info! */
        switch (zo->opcode) {
                case ZEND_JMP:
                        CALLBACK_ADD_LONG("JMP",zo->op1.u.opline_num);  
                        break;
                case ZEND_JMPZ:
                        CALLBACK_ADD_LONG("JMP",zo->op2.u.opline_num);  
                        break;
        }
        
        CALLBACK_END();
        
	 SERIALIZE_SCALAR(zo->lineno, uint);
}

void apc_deserialize_zend_op(zend_op* zo TSRMLS_DC)
{
	
        DESERIALIZE_SCALAR(&zo->opcode, zend_uchar);
        BCOMPILER_DEBUG(("Outputting %s\n", getOpcodeName(zo->opcode)));
	apc_deserialize_znode(&zo->result TSRMLS_CC);
	apc_deserialize_znode(&zo->op1 TSRMLS_CC);
	apc_deserialize_znode(&zo->op2 TSRMLS_CC);
	DESERIALIZE_SCALAR(&zo->extended_value, ulong);
	DESERIALIZE_SCALAR(&zo->lineno, uint);
}


/* type: zend_op_array */



void apc_serialize_zend_op_array(zend_op_array* zoa TSRMLS_DC)
{
	char exists;
	int i;
        CALLBACK_START(ZEND_OPCODE_ARRAY);
        CALLBACK_ADD_LONG("TYPE",(long) zoa->type);
        CALLBACK_ADD_LONG("LINE",-1);
        CALLBACK_ADD_LONG("return_reference",zoa->return_reference);
        CALLBACK_ADD_KEY_STRING("FUNCTION_NAME",zoa->function_name, strlen(zoa->function_name));  
        
        
	SERIALIZE_SCALAR(zoa->type, zend_uchar);
	apc_serialize_arg_types(zoa->arg_types TSRMLS_CC);
	apc_serialize_string(zoa->function_name TSRMLS_CC);
	SERIALIZE_SCALAR(zoa->refcount[0], zend_uint);
	SERIALIZE_SCALAR(zoa->last, zend_uint);
	SERIALIZE_SCALAR(zoa->size, zend_uint);

	/* If a file 'A' is included twice in a single request, the following 
	 * situation can occur: A is deserialized and its functions added to
	 * the global function table. On its next call, A is expired (either
	 * forcibly removed or removed due to an expired ttl). Now when A is
	 * compiled, its functions can't be added to the global function_table
	 * (they are already present) so they are serialized as an opcode
	 * ZEND_DECLARE_FUNCTION_OR_CLASS. This means that the functions will
	 * be declared at execution time. Since they are present in the global
	 * function_table, they will will also be serialized. This will cause
	 * a fatal 'failed to redclare....' error.  We avoid this by simulating
	 * the action of the parser and changing all
	 * ZEND_DECLARE_FUNCTION_OR_CLASS opcodes to ZEND_NOPs. */ 
        CALLBACK_ADD_LONG("T",(long) zoa->T);  
        CALLBACK_ADD_LONG("USES_GLOBALS",(long) zoa->uses_globals);  
        CALLBACK_END();
	for (i = 0; i < (int) zoa->last; i++) {
              
		apc_serialize_zend_op(i, &zoa->opcodes[i] TSRMLS_CC);
	}

	SERIALIZE_SCALAR(zoa->T, zend_uint);
	SERIALIZE_SCALAR(zoa->last_brk_cont, zend_uint);
	SERIALIZE_SCALAR(zoa->current_brk_cont, zend_uint);
	SERIALIZE_SCALAR(zoa->uses_globals, zend_bool);
	exists = (zoa->brk_cont_array != NULL) ? 1 : 0;
	SERIALIZE_SCALAR(exists, char);
	if (exists) {
		STORE_BYTES(zoa->brk_cont_array, zoa->last_brk_cont *
			sizeof(zend_brk_cont_element));
	}
	apc_serialize_hashtable(zoa->static_variables, apc_serialize_zval_ptr TSRMLS_CC);
	assert(zoa->start_op == NULL);
	SERIALIZE_SCALAR(zoa->return_reference, zend_bool);
	SERIALIZE_SCALAR(zoa->done_pass_two, zend_bool);
	apc_serialize_string(zoa->filename TSRMLS_CC);
        
        CALLBACK_START(ZEND_OPCODE_ARRAY);
        CALLBACK_ADD_LONG("ISEND",1);
        CALLBACK_ADD_LONG("LINE",-1);
        
        CALLBACK_END();
        
	/* zend_op_array.reserved is not used */
}

void apc_deserialize_zend_op_array(zend_op_array* zoa, int master TSRMLS_DC)
{
	char *fname;
	char exists;
	int i;

	DESERIALIZE_SCALAR(&zoa->type, zend_uchar);
	apc_create_arg_types(&zoa->arg_types TSRMLS_CC);
	apc_create_string(&zoa->function_name TSRMLS_CC);
	zoa->refcount = (int*) emalloc(sizeof(zend_uint));
	DESERIALIZE_SCALAR(zoa->refcount, zend_uint);
	DESERIALIZE_SCALAR(&zoa->last, zend_uint);
	DESERIALIZE_SCALAR(&zoa->size, zend_uint);

	zoa->opcodes = NULL;

	if (zoa->last > 0) {
		zoa->opcodes = (zend_op*) emalloc(zoa->last * sizeof(zend_op));

		for (i = 0; i < (int) zoa->last; i++) {
			apc_deserialize_zend_op(&zoa->opcodes[i] TSRMLS_CC);
			if(zoa->opcodes[i].opcode == ZEND_DECLARE_FUNCTION_OR_CLASS) {
				HashTable* table;
				char* op2str;	/* op2str and op2len are for convenience */
				int op2len;

				exists = 1;
				
				op2str = zoa->opcodes[i].op2.u.constant.value.str.val;
				op2len = zoa->opcodes[i].op2.u.constant.value.str.len;

				switch(zoa->opcodes[i].extended_value) {
				  /* a run-time function declaration */
				  case ZEND_DECLARE_FUNCTION: {
					zend_function* function;

					table = CG(function_table);

					if (zend_hash_find(table, op2str, op2len + 1,
						(void**) &function) == SUCCESS) 
					{
						zval_dtor(&zoa->opcodes[i].op1.u.constant);
						zval_dtor(&zoa->opcodes[i].op2.u.constant);
						zoa->opcodes[i].opcode = ZEND_NOP;
						memset(&zoa->opcodes[i].op1, 0, sizeof(znode));
						memset(&zoa->opcodes[i].op2, 0, sizeof(znode));
						zoa->opcodes[i].op1.op_type = IS_UNUSED;
						zoa->opcodes[i].op2.op_type = IS_UNUSED;
					}
				  } break;
				  
				  /* a run-time class declaration */
				  case ZEND_DECLARE_CLASS: {
					zend_class_entry *ce;

					table = CG(class_table);

					if (zend_hash_find(table, op2str, op2len + 1,
						(void**) &ce) == SUCCESS)
					{
						zval_dtor(&zoa->opcodes[i].op1.u.constant);
						zval_dtor(&zoa->opcodes[i].op2.u.constant);
						zoa->opcodes[i].opcode = ZEND_NOP;
						memset(&zoa->opcodes[i].op1, 0, sizeof(znode));
						memset(&zoa->opcodes[i].op2, 0, sizeof(znode));
						zoa->opcodes[i].op1.op_type = IS_UNUSED;
						zoa->opcodes[i].op2.op_type = IS_UNUSED;
					}
				  } break;
				
				  /* a run-time derived class declaration */
				  case ZEND_DECLARE_INHERITED_CLASS: {
					zend_class_entry *ce;
					char* class_name;
					char* parent_name;
					int class_name_length;
					namearray_t* children;
					int minSize;

					table = CG(class_table);

					/* op2str is a class name of the form "base:derived",
					 * where derived is the class being declared and base
					 * is its parent in the class hierarchy. Extract the
					 * two names into parent_name and class_name. */
					
					parent_name = apclib_estrdup(op2str);
					if ((class_name = strchr(parent_name, ':')) == 0) {
						zend_error(E_CORE_ERROR,"Invalid runtime class entry");
					}
					*class_name++ = '\0';	/* advance past ':' */

					if (zend_hash_find(table, class_name, strlen(class_name)+1,
						(void**) &ce) == SUCCESS) 
					{
						zval_dtor(&zoa->opcodes[i].op1.u.constant);
						zval_dtor(&zoa->opcodes[i].op2.u.constant);
						zoa->opcodes[i].opcode = ZEND_NOP;
						memset(&zoa->opcodes[i].op1, 0, sizeof(znode));
						memset(&zoa->opcodes[i].op2, 0, sizeof(znode));
						zoa->opcodes[i].op1.op_type = IS_UNUSED;
						zoa->opcodes[i].op2.op_type = IS_UNUSED;
					}

					/* The parent class hasn't been compiled yet, most likely
					 * because it is defined in an included file. We must defer
					 * this inheritance until the parent class is created. Add
					 * a tuple for this class and its parent to the deferred
					 * inheritance table */
		
					class_name_length = strlen(class_name);
					//children = apc_nametable_retrieve(deferred_inheritance,
					//                                  parent_name TSRMLS_CC);
		
					if (children == 0) {
						/* Create and initialize a new namearray_t for this
						 * class's parent and insert into the deferred
						 * inheritance table. */
		
						children = (namearray_t*) malloc(sizeof(namearray_t));
						children->length = 0;
						children->size = class_name_length + 1;
						children->strings = (char*) malloc(children->size);
		
						//apc_nametable_insert(deferred_inheritance,
						//                     parent_name, children TSRMLS_CC);
					}

					/* a deferred class can't be a top-level parent */
					//apc_nametable_insert(non_inheritors, class_name, NULL TSRMLS_CC);


					minSize = children->length + class_name_length + 1;
		
					if (minSize > children->size) {
						/* The strings array (children->strings) is not big
						 * enough to store this class name. Expand the array
						 * using an exponential resize. */

						while (minSize > children->size) {
							children->size *= 2;
						}
						children->strings = apclib_erealloc(children->strings,
						                                 children->size);
					}
					memcpy(children->strings + children->length,
						class_name, class_name_length + 1);
					children->length += class_name_length + 1;
				  } break;

				  default:
					break;
				}
			}
		}
	}

	DESERIALIZE_SCALAR(&zoa->T, zend_uint);
	DESERIALIZE_SCALAR(&zoa->last_brk_cont, zend_uint);
	DESERIALIZE_SCALAR(&zoa->current_brk_cont, zend_uint);
	DESERIALIZE_SCALAR(&zoa->uses_globals, zend_bool);
	DESERIALIZE_SCALAR(&exists, char);
	zoa->brk_cont_array = NULL;
	if (exists) {
		zoa->brk_cont_array = (zend_brk_cont_element*)
			emalloc(zoa->last_brk_cont * sizeof(zend_brk_cont_element));
		LOAD_BYTES(zoa->brk_cont_array, zoa->last_brk_cont *
			sizeof(zend_brk_cont_element));
	}
	apc_create_hashtable(&zoa->static_variables, apc_create_zval, NULL, sizeof(zval *) TSRMLS_CC);
 
	zoa->start_op = NULL;
 
	DESERIALIZE_SCALAR(&zoa->return_reference, zend_bool);
	DESERIALIZE_SCALAR(&zoa->done_pass_two, zend_bool);
	apc_create_string(&fname TSRMLS_CC);
	zoa->filename = zend_set_compiled_filename(fname TSRMLS_CC);
	apclib_free_string(&fname);
	if(master) {
		// apc_nametable_difference(parental_inheritors, non_inheritors TSRMLS_CC);
		// apc_nametable_apply(parental_inheritors, call_inherit TSRMLS_CC);
	}
}

void apc_create_zend_op_array(zend_op_array** zoa TSRMLS_DC)
{
	*zoa = (zend_op_array*) emalloc(sizeof(zend_op_array));
	apc_deserialize_zend_op_array(*zoa, 0 TSRMLS_CC);
}


/* type: zend_internal_function */


	
void apc_serialize_zend_internal_function(zend_internal_function* zif TSRMLS_DC)
{
	SERIALIZE_SCALAR(zif->type, zend_uchar);
	apc_serialize_arg_types(zif->arg_types TSRMLS_CC);
	apc_serialize_string(zif->function_name TSRMLS_CC);
	SERIALIZE_SCALAR(zif->handler, void*);	
}

void apc_deserialize_zend_internal_function(zend_internal_function* zif TSRMLS_DC)
{
	DESERIALIZE_SCALAR(&zif->type, zend_uchar);
	apc_create_arg_types(&zif->arg_types TSRMLS_CC);
	apc_create_string(&zif->function_name TSRMLS_CC);
	DESERIALIZE_SCALAR(&zif->handler, void*);
}


/* type: zend_overloaded_function */


void apc_serialize_zend_overloaded_function(zend_overloaded_function* zof TSRMLS_DC)
{
	SERIALIZE_SCALAR(zof->type, zend_uchar);
	apc_serialize_arg_types(zof->arg_types TSRMLS_CC);
	apc_serialize_string(zof->function_name TSRMLS_CC);
	SERIALIZE_SCALAR(zof->var, zend_uint);	
}

void apc_deserialize_zend_overloaded_function(zend_overloaded_function* zof TSRMLS_DC)
{
	DESERIALIZE_SCALAR(&zof->type, zend_uchar);
	apc_create_arg_types(&zof->arg_types TSRMLS_CC);
	apc_create_string(&zof->function_name TSRMLS_CC);
	DESERIALIZE_SCALAR(&zof->var, zend_uint);
}


/* type: zend_function */

void apc_serialize_zend_function(zend_function* zf TSRMLS_DC)
{
	/* zend_function come in 4 different types with different internal 
	 * structs. */    
        

	BCOMPILER_DEBUG(("start serialize zend function type:%d\n", zf->type));
        SERIALIZE_SCALAR(zf->type, zend_uchar);
	BCOMPILER_DEBUG(("start serialize zend function\n"));
         
	switch(zf->type) {
	  case ZEND_INTERNAL_FUNCTION:
		apc_serialize_zend_internal_function(&zf->internal_function TSRMLS_CC);
		break;
	  case ZEND_OVERLOADED_FUNCTION:
		apc_serialize_zend_overloaded_function(&zf->overloaded_function TSRMLS_CC);
		break;
	  case ZEND_USER_FUNCTION:
	  case ZEND_EVAL_CODE:
                BCOMPILER_DEBUG(("start serialize op array\n"));
		apc_serialize_zend_op_array(&zf->op_array TSRMLS_CC);
		break;
	  default:
		/* the above are all valid zend_function types.  If we hit this
		 * case something has gone very very wrong. */
		assert(0);
	}
}

void apc_deserialize_zend_function(zend_function* zf TSRMLS_DC)
{
	/* PEEK_SCALAR(&zf->type, zend_uchar); */
        DESERIALIZE_SCALAR(&zf->type, zend_uchar);
	switch(zf->type) {
	  case ZEND_INTERNAL_FUNCTION:
		apc_deserialize_zend_internal_function(&zf->internal_function TSRMLS_CC);
		break;
      case ZEND_OVERLOADED_FUNCTION:
		apc_deserialize_zend_overloaded_function(&zf->overloaded_function TSRMLS_CC);
		break;
      case ZEND_USER_FUNCTION:
      case ZEND_EVAL_CODE:
		apc_deserialize_zend_op_array(&zf->op_array, 0 TSRMLS_CC);
		break;
	  default:
		/* the above are all valid zend_function types.  If we hit this
         * case something has gone very very wrong. */
		assert(0);
	}
}

void apc_create_zend_function(zend_function** zf TSRMLS_DC)
{
	*zf = (zend_function*) emalloc(sizeof(zend_function));
	apc_deserialize_zend_function(*zf TSRMLS_CC);
}
void apc_free_zend_function(zend_function** zf TSRMLS_DC)
{
  if (*zf != NULL)
    efree(*zf);
  *zf = NULL;
}


/* The logic for serialize_class_table and deserialize_class_table
 * mirror that of the corresponding function_table functions (see
 * above), with classes in place of functions. */
 


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
